/*************************************************************************
 *
 *                  Elektor Car Accelerometer Project V1.0
 *
 *************************************************************************
 * FileName:        main.c
 * Dependencies:    main.h
 * Processor:       HCS08
 * Compiler:        Code Warrior 5.1
 * Company:         Freescale Semiconductor
 *
 * Author                Date          Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * James Courtier        00/00/2006    Original.
 ************************************************************************/


/** I N C L U D E S *****************************************************/
#include <hidef.h>      /* for EnableInterrupts macro */
#include "derivative.h" /* include peripheral declarations */
#include "main.h"
#include <stdio.h>

/** G L O B A L   V A R I A B L E S *************************************/
char PowerStat=ON;    //Power status set to on
char LEDMag[4]={0};   //LED Matrix bitmap
int timer;            //Global timer incremented by Modulo Timer Interupt
signed char cal_x=0;    //Offset for X value
signed char cal_y=0;    //Offset for Y value


/** G L O B A L   C O N S T A N T S *************************************/
/* Accelerometer calibration data */
/* Vechicle used for cal - Vauxhall Corsa Design 1.2i 16V Twinport */
const char PosR[3]={160,210,225}; //Max positive value
const char PosA[3]={150,172,195}; //Mid positive value
const char PosG[3]={140,152,165}; //Min positive value
const char NegG[3]={120,105, 90}; //Min negative value
const char NegA[3]={110, 85, 60}; //Mid negative value
const char NegR[3]={100, 65, 30}; //Max negative value
//                    |   |   |
//                    |   |   +-----Low sensitivity
//                    |   +---------Mid sensitivity
//                    +-------------Max sensitivity

/** I N T E R U P T   S E R V I C E   R O U T I N E S *******************/
interrupt 12 void MTIM_ISR(void){
  /* Increment global timer */
  timer++;
  /* Find out which column is currently lit */
  switch (PTBD & 0b01111000){
    
    case LEDCOLF : PTBD = LEDCOLB | LEDMag[NegYLED]; break;      
    case LEDCOLB : PTBD = LEDCOLL | LEDMag[PosXLED]; break;      
    case LEDCOLL : PTBD = LEDCOLR | LEDMag[NegXLED]; break;      
    case LEDCOLR : PTBD = LEDCOLF | LEDMag[PosYLED]; break;      
    default      : PTBD = LEDCOLF | LEDMag[PosYLED]; break;   
  }   
  MTIMSC_TOF = 0;  //Clear Timer0 interupt flag
}

interrupt 23 void VRTI_ISR(void){
  if (buttonstat() == PwrBut) PowerStat = ON; //If Power button is pressed set status to On
  SRTISC_RTIACK = 1; //RTI flag
}

/** P U B L I C   F U N C T I O N S *************************************/

//***********************************************************************
//** Function:      MAIN
//** Description:   User code entry point - Hub for calling other functions
//** Parameters:    None
//** Returns:       None (Should never exit)
//***********************************************************************
void main(void) {
  char Sensitivity=HighSen; //Default sensitivity set to low

  EnableInterrupts;         //Global interupt enable
  init();                   //Initialise MCU registers
  PowerOff();               //Switch off
  AccCal();                 //Zero the acellerometer
  Sensitivity = IncSens(HighSen); //Set initial sensitivity
  for(;;) {         
    switch(buttonstat()){   //Check to see if a button has been pressed
      case (PwrBut) :       //Power button pushed?
        DisplayAnim(OFF);   //Do off animation
        PowerOff();         //Change power mode  
        AccCal();           //Zero the acellerometer
        Sensitivity = IncSens(HighSen); //Set sensitivity         
      break;
    
      case (SenBut) :       //Sensitivity pushed?
        Sensitivity = IncSens(Sensitivity); 
      break;
      
      default :             //No buttons currently pressed?
        UpdateDisplay(Sensitivity);    //Update display with new acellerometer values
      break;
    }
  }
}

//***********************************************************************
//** Function:      init
//** Description:   Initialises MCU registers
//** Parameters:    None
//** Returns:       None
//***********************************************************************
void init(void) {
  /* Setup interupts */
  SOPT1   = 0b00100010; //Disable watchdog & Enable STOP instruction
  SRTISC  = 0b00000110; //Real time interupt set to 512ms (internal clock)

  /* Setup pins */
  PTBD    = 0b00000000; //Disable all the columns and switch off LEDs
  PTBDD   = 0b11111111; //Pins 0->7 outputs
  PTBDS   = 0b01111111; //Pins 0->6 High drive enable
  PTBPE   = 0b00000111; //Pins 0->2 Resistor pull up enable     
  PTBSE   = 0b00000000; //Turn off PORTB slew control
  APCTL1  = 0b00000111; //Define PORTA 0->2 as analogue inputs
  
  /* Setup timer */
  MTIMCLK = 0b00000100; //Timer uses bus clock and 16scaler
  MTIMSC  = 0b00100000; //Reset timer
  MTIMSC  = 0b01000000; //Enable timer interupt

  /* Setup ADC */
  PTAD    = 0b00000000;
  PTADD   = 0b00001000;
}

//***********************************************************************
//** Function:      AccCal
//** Description:   Computes offset required for accelerometer
//** Parameters:    None
//** Returns:       None
//***********************************************************************
void AccCal(void){
  cal_x = AccFil(ACC_X) - AccCen;
  cal_y = AccFil(ACC_Y) - AccCen;  
}

//***********************************************************************
//** Function:      Byte2LED
//** Description:   Converts a byte to a scale on the display
//** Parameters:    char Val - Value to be displayed
//**                char Dir - Scale to be updated: DirX or DirY
//**                char SenMode - Sensitivity (High,Med,Low)Sen
//** Returns:       None
//***********************************************************************
void Byte2LED(char Val, char Dir, char SenMode){ 
  if (Val > PosG[SenMode]) {
    LEDMag[NegYLED + Dir] = LEDOFF;
    if (Val > PosR[SenMode]) LEDMag[PosYLED + Dir] = LEDGAR;
    else {
      if (Val > PosA[SenMode]) LEDMag[PosYLED + Dir] = LEDGA;
      else {
        LEDMag[PosYLED + Dir] = LEDG;
      }   
    }
  } 
  else{   
    if (Val < NegG[SenMode]) {
      LEDMag[PosYLED + Dir] = LEDOFF;   
      if (Val < NegR[SenMode]) LEDMag[NegYLED + Dir] = LEDGAR;
      else{
        if (Val < NegA[SenMode]) LEDMag[NegYLED + Dir] = LEDGA;
        else{
          LEDMag[NegYLED + Dir] = LEDG;   
        }
      }
    }        
    else{
      LEDMag[PosYLED + Dir] = LEDOFF;               
      LEDMag[NegYLED + Dir] = LEDOFF;
    }
  }
}

//***********************************************************************
//** Function:      PowerOff
//** Description:   Shuts down the MCU, and polls the Power button
//** Parameters:    None
//** Returns:       None
//***********************************************************************
void PowerOff(void){
  LEDPRTD     = 0x00;     //Turn off all LEDs
  MTIMSC      = 0x00;     //Turn off LED multiplexing    
  SRTISC_RTIE = 1;        //Enable RTI
  PowerStat   = OFF; 
  PTAD        = 0x00;     //Disable acellerometer  
  while (PowerStat == OFF){ //Stay in sleep while the app is "Off"
    asm("STOP");            //Sleep (Good night!) 
  }              
  PTAD        = ACC_SLP;    //Enable acellerometer
  SRTISC_RTIE = 0b00000000; //Disable RTI       
  MTIMSC      = 0b01000000; //Turn on LED multiplexing
  DisplayAnim(ON);          //Do start-up animation
}

//***********************************************************************
//** Function:      IncSens
//** Description:   Increases current sensitivity when called
//** Parameters:    char Sens - current sensitivity
//** Returns:       char - New sensitivity
//***********************************************************************
char IncSens(char Sens){
  /* Switch the current sensitivity */
  switch(Sens){
    case(HighSen) :
      /* Increment sensitivity */
      Sens = LowSen;          
      /* Switch on all green leds */
      LEDMag[PosYLED] = LEDG; //
      LEDMag[PosXLED] = LEDG;
      LEDMag[NegYLED] = LEDG;
      LEDMag[NegXLED] = LEDG;      
    break;
  
    case(MedSen) :
      /* Increment sensitivity */
      Sens = HighSen;
      /* Switch on all red leds */
      LEDMag[PosYLED] = LEDR;
      LEDMag[PosXLED] = LEDR;
      LEDMag[NegYLED] = LEDR;
      LEDMag[NegXLED] = LEDR;      
    break;
  
    case(LowSen) :
      /* Increment sensitivity */
      Sens = MedSen;
      /* Switch on all amber leds */
      LEDMag[PosYLED] = LEDA;
      LEDMag[PosXLED] = LEDA;
      LEDMag[NegYLED] = LEDA;
      LEDMag[NegXLED] = LEDA;      
    break;
    
    default :
      /* Increment sensitivity */
      Sens = LowSen;
      /* Switch on all green leds */
      LEDMag[PosYLED] = LEDG;
      LEDMag[PosXLED] = LEDG;
      LEDMag[NegYLED] = LEDG;
      LEDMag[NegXLED] = LEDG;      
    break;
  }
  /* Debounce and sustain delay */
  delay(1000);
  /* Return the new sensivity value */
  return(Sens);
}

//***********************************************************************
//** Function:      UpdateDisplay
//** Description:   Reads accelerometer x & y val, filters, and updates LEDs
//** Parameters:    char Sens - Sensitivity
//** Returns:       None
//***********************************************************************
void UpdateDisplay(char Sens){
  /* Get Accelerometer value and move it into sample window */
  int acc_x = (int)AccFil(ACC_X) - (signed char)cal_x;
  int acc_y = (int)AccFil(ACC_Y) - (signed char)cal_y;
  
  /* Check to make sure value isn't too big or small for a char */
  if (acc_x > 255) acc_x = 255;
  if (acc_y > 255) acc_y = 255;
  if (acc_x < 0) acc_x = 0;
  if (acc_y < 0) acc_y = 0;
     
  /* Update display with new X and Y data */
  Byte2LED(acc_x&0xFF, DirX, Sens);
  Byte2LED(acc_y&0xFF, DirY, Sens);  
}

//***********************************************************************
//** Function:      AccFil
//** Description:   Reads accelerometer and applies a short averaging filter
//** Parameters:    char acc - ADC channel to read
//** Returns:       char - Average value
//***********************************************************************
char AccFil(char acc){
  int temp=0;
  char j;
  for (j=0 ; j<32 ; j++){
    delay(1); 
    temp += readADC(acc);
  }
  return((temp/32)&0xFF);
}


//***********************************************************************
//** Function:      DisplayAnim
//** Description:   Manipulates LEDMag variables to produce on&off anims
//** Parameters:    char Mode - On or Off animation (On&Off defined in header)
//** Returns:       None
//***********************************************************************
void DisplayAnim(char mode){
  char j;
  if (mode == OFF) {
    /* Clear display */
    LEDMag[PosYLED] = LEDOFF;
    LEDMag[PosXLED] = LEDOFF;
    LEDMag[NegYLED] = LEDOFF;
    LEDMag[NegXLED] = LEDOFF; 
    
    /* Do line-crossover animation */
    for (j=128 ; j<255 ; j++){
      Byte2LED(j, DirY, HighSen);
      delay(3);
    }
    for (j=128 ; j<255 ; j++){
      Byte2LED(j, DirX, HighSen);
      delay(3);
    }    
    for (j=255 ; j>1 ; j--){
      Byte2LED(j, DirY, HighSen);
      delay(3);
    }   
    for (j=255 ; j>1 ; j--){
      Byte2LED(j, DirX, HighSen);
      delay(3);
    }
    for (j=0 ; j<128 ; j++){
      Byte2LED(j, DirY, HighSen);
      delay(3);
    }
    for (j=0 ; j<128 ; j++){
      Byte2LED(j, DirX, HighSen);
      delay(3);
    }
  }
  else{
    /* Flash all LED's On and Off twice */
    for (j=0 ; j< 2 ; j++){
      LEDMag[PosYLED] = LEDGAR;
      LEDMag[PosXLED] = LEDGAR;
      LEDMag[NegYLED] = LEDGAR;
      LEDMag[NegXLED] = LEDGAR;      
      delay(500);
      LEDMag[PosYLED] = LEDOFF;
      LEDMag[PosXLED] = LEDOFF;
      LEDMag[NegYLED] = LEDOFF;
      LEDMag[NegXLED] = LEDOFF;      
      delay(500);
    }
  }
}

//***********************************************************************
//** Function:      BUTTONSTAT
//** Description:   Reads ADC to discover which button is pressed
//** Parameters:    None
//** Returns:       Returns current button pressed (1 or 2), 0 if none.
//***********************************************************************
char buttonstat(void){ 
  unsigned int averageval=0;
  unsigned int temp;
  unsigned char n;
  
  /* Compute average value of button resistor ladder */ 
  for (n=0;n<avsamples;n++){    
      averageval += readADC(BUTTONS);
      for(temp = 0 ; temp < 1000 ; temp++) asm("NOP");
  }
  n = ((unsigned int)averageval/(unsigned char)avsamples )&255;
  /* Decode which button was pressed based on ADC value */
  if ((but1min <= n) && (n <= but1max)) return(1);
  if ((but2min <= n) && (n <= but2max)) return(2);
  return(0);
}

//***********************************************************************
//** Function:      Delay
//** Description:   Simple loop delay, using variable timer incremented
//**                by MTIM interupt.
//** Parameters:    int n - Number of times to loop
//** Returns:       None
//***********************************************************************
void delay(int n){
  timer=0;
  while (timer < n);
}

//***********************************************************************
//** Function:      readADC
//** Description:   Reads current ADC value for a selected channel
//** Parameters:    char chan - Channel to read
//** Returns:       Char - current ADC value for selected channel
//***********************************************************************
char readADC(char chan){
    ADCSC1_ADCH = chan;     //Set ADC channel to chan
    while(!ADCSC1_COCO);    //Wait until conversion complete
    return(ADCRL);
}
